/*
 * Copyright (C) 2009 Red Hat, Inc.
 *
 * This file is released under the GPL.
 */
#include <stdlib.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>
#include "linux/dm-clog-tfr.h"
#include "logging.h"
#include "compat.h"

#define DM_CLOG_RESPONSE 0x1000 /* in last byte of 32-bit value */
#define DM_CLOG_CHECKPOINT_READY 21
#define DM_CLOG_MEMBER_JOIN      22

/* Version 1 structures never were public */

/*************************************************************************
 ** RHEL5.2 STRUCTURES
 ************************************************************************/
struct clog_tfr_v2 {
	uint64_t private[2];
	char uuid[DM_UUID_LEN]; /* Ties a request to a specific mirror log */

	int error;              /* Used by server to inform of errors */
	uint32_t originator;    /* Cluster ID of this machine */

	uint64_t seq;           /* Sequence number for request */
	uint32_t request_type;  /* DM_CLOG_* */
	uint32_t data_size;     /* How much data (not including this struct) */
	char data[0];
};


/*************************************************************************
 ** RHEL5.3 STRUCTURES
 ************************************************************************/
struct clog_tfr_v3 {
	uint64_t private[2];
	char uuid[DM_UUID_LEN]; /* Ties a request to a specific mirror log */

	int error;              /* Used by server to inform of errors */
	uint32_t originator;    /* Cluster ID of this machine */

	uint32_t seq;           /* Sequence number for request */
	uint32_t request_type;  /* DM_CLOG_* */
	uint32_t data_size;     /* How much data (not including this struct) */
	char data[0];
};


/*************************************************************************
 ** RHEL5.4+ STRUCTURES
 ************************************************************************/
struct clog_tfr_v4 {
	uint64_t private[2];
	char uuid[DM_UUID_LEN]; /* Ties a request to a specific mirror log */
	char uuid_padding[3];   /* DM_UUID_LEN == 129 */
	uint32_t uuid_instance; /* Allows more than one instance w/ same uuid */

	int32_t error;              /* Used by server to inform of errors */
	uint32_t originator;    /* Cluster ID of this machine */

	uint32_t seq;           /* Sequence number for request */
	uint32_t request_type;  /* DM_CLOG_* */
	uint64_t data_size;     /* How much data (not including this struct) */
	char data[0];
};


/*************************************************************************
 ** UPSTREAM STRUCTURES	
 ************************************************************************/

/*
 * The upstream machine <-> machine communication structure
 */
struct clog_version {
	uint64_t version;
	uint64_t flags;
};

struct compat_list {
	struct compat_list *a, *b;
};

/*
 * The upstream (2.6.32) kernel <-> user communication structure
 */
struct dm_ulog_request {
        uint64_t luid;
        char uuid[DM_UUID_LEN];
        char padding[7];        /* Padding because DM_UUID_LEN = 129 */

        int32_t error;          /* Used to report back processing errors */

        uint32_t seq;           /* Sequence number for request */
        uint32_t request_type;  /* DM_ULOG_* defined above */
        uint32_t data_size;     /* How much data (not including this struct) */

        char data[0];
};

struct clog_request {
	union {
		struct compat_list list;
		struct clog_version v;
	} u;

        /*
         * 'originator' is the machine from which the requests
         * was made.
         */
        uint32_t originator;

        /*
         * 'pit_server' is the "point-in-time" server for the
         * request.  (I.e.  The machine that was the server at
         * the time the request was issued - only important during
         * startup.
         */
        uint32_t pit_server;

        /*
         * The request from the kernel that is being processed
         */
        struct dm_ulog_request u_rq;
};

#define COMPAT_SIZE   512
#define COMPAT_OFFSET 256
unsigned char buffer[COMPAT_SIZE];

static int v2_to_v5(struct clog_tfr_v2 *v2, struct clog_request *rq)
{
	LOG_ERROR("Unable to add compat layer v2->v5");
	return -EINVAL;
}

static int v3_to_v5(struct clog_tfr_v3 *v3, struct clog_request *rq)
{
	LOG_ERROR("Unable to add compat layer v3->v5");
	return -EINVAL;
}

static int v4_to_v5_copy_data(struct clog_tfr_v4 *v4, struct clog_request *rq)
{
	int data_size = 0;
	uint32_t rq_type = v4->request_type & ~DM_CLOG_RESPONSE;
	
	if (v4->request_type & DM_CLOG_RESPONSE) {
		switch (rq_type) {
		case DM_CLOG_CTR:
		case DM_CLOG_DTR:
		case DM_CLOG_PRESUSPEND:
		case DM_CLOG_POSTSUSPEND:
		case DM_CLOG_RESUME:
		case DM_CLOG_FLUSH:
		case DM_CLOG_MARK_REGION:
		case DM_CLOG_CLEAR_REGION:
		case DM_CLOG_SET_REGION_SYNC:
		case DM_CLOG_CHECKPOINT_READY:
		case DM_CLOG_MEMBER_JOIN:
			/* No outbound data */
			break;

		case DM_CLOG_GET_REGION_SIZE:
			data_size = sizeof(uint64_t);
			if (rq) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)v4->data;
				b64 = (uint64_t *)rq->u_rq.data;
				*b64 = cpu_to_le64(*a64);
			}
			break;
		case DM_CLOG_IS_CLEAN:
		case DM_CLOG_IN_SYNC:
			data_size = sizeof(int64_t);
			if (rq) {
				int i;
				int64_t *p64;

				i = *((int *)v4->data);
				p64 = (int64_t *)rq->u_rq.data;
				*p64 = cpu_to_le64((int64_t)i);
			}
			break;
		case DM_CLOG_GET_RESYNC_WORK:
			data_size = sizeof(uint64_t) * 2;
			if (rq) {
				int32_t i;
				int64_t *pi64;
				uint64_t *pu64, u64;

				i = *((int32_t *)v4->data);
				u64 = ((uint64_t *)v4->data)[1];
				pi64 = (int64_t *)rq->u_rq.data;
				pu64 = (uint64_t *)rq->u_rq.data;
				pu64++;
				*pi64 = cpu_to_le64((int64_t )i);
				*pu64 = cpu_to_le64(u64);
			}
			break;
		case DM_CLOG_GET_SYNC_COUNT:
			data_size = sizeof(uint64_t);
			if (rq) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)v4->data;
				b64 = (uint64_t *)rq->u_rq.data;
				*b64 = cpu_to_le64(*a64);
			}
			break;			
		case DM_CLOG_STATUS_INFO:
		case DM_CLOG_STATUS_TABLE:
			data_size = v4->data_size;
			if (rq)
				memcpy(rq->u_rq.data, v4->data, v4->data_size);
			break;
		case DM_CLOG_IS_REMOTE_RECOVERING:
			data_size = sizeof(int32_t) +
				sizeof(uint32_t) + sizeof(uint64_t);
			if (rq) {
				int32_t is_recovering;
				uint64_t in_sync_hint, *pu64;
				int64_t *pi64;

				is_recovering = *((int32_t *)v4->data);
				in_sync_hint = ((uint64_t *)v4->data)[1];
				pi64 = (int64_t *)rq->u_rq.data;
				pu64 = ((uint64_t *)rq->u_rq.data) + 1;
				*pi64 = cpu_to_le64((int64_t)is_recovering);
				*pu64 = cpu_to_le64(in_sync_hint);
			}
			break;
		default:
			LOG_ERROR("Unknown request type, %u", rq_type);
			return -EINVAL;
		}
	} else {
		switch (rq_type) {
		case DM_CLOG_CTR:
		case DM_CLOG_DTR:
		case DM_CLOG_PRESUSPEND:
		case DM_CLOG_POSTSUSPEND:
		case DM_CLOG_RESUME:
		case DM_CLOG_GET_REGION_SIZE:
		case DM_CLOG_FLUSH:
		case DM_CLOG_GET_RESYNC_WORK:
		case DM_CLOG_GET_SYNC_COUNT:
		case DM_CLOG_STATUS_INFO:
		case DM_CLOG_STATUS_TABLE:
		case DM_CLOG_CHECKPOINT_READY:
		case DM_CLOG_MEMBER_JOIN:
			/* No incoming data */
			break;
		case DM_CLOG_IS_CLEAN:
		case DM_CLOG_IN_SYNC:
			data_size = sizeof(uint64_t);
			if (rq) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)v4->data;
				b64 = (uint64_t *)rq->u_rq.data;
				*b64 = cpu_to_le64(*a64);
			}
			break;
		case DM_CLOG_MARK_REGION:
		case DM_CLOG_CLEAR_REGION:
			data_size = v4->data_size;
			if (rq) {
				int i, end = data_size/sizeof(uint64_t);
				uint64_t *a64, *b64;

				a64 = (uint64_t *)v4->data;
				b64 = (uint64_t *)rq->u_rq.data;
				for (i = 0; i < end; i++)
					b64[i] = cpu_to_le64(a64[i]);
			}
			break;
		case DM_CLOG_SET_REGION_SYNC:
			data_size = sizeof(uint64_t) * 2;
			if (rq) {
				uint64_t r, *pu64;
				int64_t *pi64;
				int32_t in_sync;

				r = *((uint64_t *)v4->data);
				in_sync = ((int32_t *)v4->data)[3];
				pu64 = (uint64_t *)rq->u_rq.data;
				pi64 = ((int64_t *)rq->u_rq.data) + 1;
				*pu64 = cpu_to_le64(r);
				*pi64 = cpu_to_le64((int64_t)in_sync);
			}
			break;
		case DM_CLOG_IS_REMOTE_RECOVERING:
			data_size = sizeof(uint64_t);
			if (rq) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)v4->data;
				b64 = (uint64_t *)rq->u_rq.data;
				*b64 = cpu_to_le64(*a64);
			}
			break;
		default:
			LOG_ERROR("Unknown request type, %u", rq_type);
			return -EINVAL;
		}
	}

	return data_size;
}

static int v4_to_v5(struct clog_tfr_v4 *v4, struct clog_request *rq)
{
	int size;
	struct dm_ulog_request *u_rq = &rq->u_rq;

	size = sizeof(*rq) + v4_to_v5_copy_data(v4, NULL);

	if (size > (COMPAT_SIZE - COMPAT_OFFSET)) {
		LOG_ERROR("Not enough space for compatibility layer (v4->v5)");
		return -EINVAL;
	}

	memcpy(u_rq->uuid, v4->uuid, DM_UUID_LEN);
	u_rq->error = cpu_to_le32(v4->error);
	u_rq->seq = cpu_to_le32(v4->seq);
	u_rq->request_type = cpu_to_le32(v4->request_type);
	u_rq->data_size = cpu_to_le64(v4->data_size);

	rq->originator = cpu_to_le32(v4->originator);

	v4_to_v5_copy_data(v4, rq);

	return COMPAT_OFFSET + sizeof(*rq) + u_rq->data_size;
}

#define switch_64(x, to_wire) \
	(to_wire) ? le64_to_cpu(x) : cpu_to_le64(x)

void v5_data_endian_switch(struct clog_request *rq, int to_wire)
{
	int i, end;
	int64_t *pi64;
	uint64_t *pu64;
	uint32_t rq_type = rq->u_rq.request_type & ~DM_CLOG_RESPONSE;
	
	if (rq->u_rq.request_type & DM_CLOG_RESPONSE) {
		switch (rq_type) {
		case DM_CLOG_CTR:
		case DM_CLOG_DTR:
			LOG_ERROR("Invalid response type in endian switch");
			exit(EXIT_FAILURE);

		case DM_CLOG_PRESUSPEND:
		case DM_CLOG_POSTSUSPEND:
		case DM_CLOG_RESUME:
		case DM_CLOG_FLUSH:
		case DM_CLOG_MARK_REGION:
		case DM_CLOG_CLEAR_REGION:
		case DM_CLOG_SET_REGION_SYNC:
		case DM_CLOG_CHECKPOINT_READY:
		case DM_CLOG_MEMBER_JOIN:
		case DM_CLOG_STATUS_INFO:
		case DM_CLOG_STATUS_TABLE:
			/* No outbound data */
			break;

		case DM_CLOG_GET_REGION_SIZE:
		case DM_CLOG_GET_SYNC_COUNT:
			pu64 = (uint64_t *)rq->u_rq.data;
			*pu64 = switch_64(*pu64, to_wire);
			break;
		case DM_CLOG_IS_CLEAN:
		case DM_CLOG_IN_SYNC:
			pi64 = (int64_t *)rq->u_rq.data;
			*pi64 = switch_64(*pi64, to_wire);
			break;
		case DM_CLOG_GET_RESYNC_WORK:
		case DM_CLOG_IS_REMOTE_RECOVERING:
			pi64 = (int64_t *)rq->u_rq.data;
			pu64 = ((uint64_t *)rq->u_rq.data) + 1;
			*pi64 = switch_64(*pi64, to_wire);
			*pu64 = switch_64(*pu64, to_wire);
			break;
		default:
			LOG_ERROR("Unknown request type, %u", rq_type);
			return;
		}
	} else {
		switch (rq_type) {
		case DM_CLOG_CTR:
		case DM_CLOG_DTR:
			LOG_ERROR("Invalid request type in endian switch");
			exit(EXIT_FAILURE);

		case DM_CLOG_PRESUSPEND:
		case DM_CLOG_POSTSUSPEND:
		case DM_CLOG_RESUME:
		case DM_CLOG_GET_REGION_SIZE:
		case DM_CLOG_FLUSH:
		case DM_CLOG_GET_RESYNC_WORK:
		case DM_CLOG_GET_SYNC_COUNT:
		case DM_CLOG_STATUS_INFO:
		case DM_CLOG_STATUS_TABLE:
		case DM_CLOG_CHECKPOINT_READY:
		case DM_CLOG_MEMBER_JOIN:
			/* No incoming data */
			break;
		case DM_CLOG_IS_CLEAN:
		case DM_CLOG_IN_SYNC:
		case DM_CLOG_IS_REMOTE_RECOVERING:
			pu64 = (uint64_t *)rq->u_rq.data;
			*pu64 = switch_64(*pu64, to_wire);
			break;
		case DM_CLOG_MARK_REGION:
		case DM_CLOG_CLEAR_REGION:
			end = rq->u_rq.data_size/sizeof(uint64_t);

			pu64 = (uint64_t *)rq->u_rq.data;
			for (i = 0; i < end; i++)
				pu64[i] = switch_64(pu64[i], to_wire);
			break;
		case DM_CLOG_SET_REGION_SYNC:
			pu64 = (uint64_t *)rq->u_rq.data;
			pi64 = ((int64_t *)rq->u_rq.data) + 1;
			*pu64 = switch_64(*pu64, to_wire);
			*pi64 = switch_64(*pi64, to_wire);
			break;
		default:
			LOG_ERROR("Unknown request type, %u", rq_type);
			exit(EXIT_FAILURE);
		}
	}
}

int v5_endian_to_wire(struct clog_request *rq)
{
	int size;
	struct dm_ulog_request *u_rq = &rq->u_rq;

	size = sizeof(*rq) + u_rq->data_size;

	u_rq->error = cpu_to_le32(u_rq->error);
	u_rq->seq = cpu_to_le32(u_rq->seq);
	u_rq->request_type = cpu_to_le32(u_rq->request_type);
	u_rq->data_size = cpu_to_le64(u_rq->data_size);

	rq->originator = cpu_to_le32(rq->originator);

	v5_data_endian_switch(rq, 1);

	return size;
}

int add_compatibility_layer(void *from, void **compat_version,
			    unsigned from_version)
{
	int from_size = COMPAT_SIZE;
	struct clog_tfr_v2 *v2 = from;
	struct clog_tfr_v3 *v3 = from;
	struct clog_tfr_v4 *v4 = from;

	switch (from_version) {
	case 2:
		from_size = sizeof(*v2) + v2->data_size;
		break;
	case 3:
		from_size = sizeof(*v3) + v3->data_size;
		break;
	case 4:
		from_size = sizeof(*v4) + v4->data_size;
		break;
	case 5:
		return v5_endian_to_wire(from);
	default:
		LOG_ERROR("Invalid 'from_version'");
		return -EINVAL;
	}

	if (from_size > COMPAT_OFFSET) {
		LOG_ERROR("Not enough space for request (v%d:%d)",
			  from_version, from_size);
		return -EINVAL;
	}

	memset(buffer, 0, COMPAT_SIZE);
	memcpy(buffer, from, from_size);
	*compat_version = buffer;

	switch (from_version) {
	case 2:
		return v2_to_v5((void *)buffer, (void *)buffer + COMPAT_OFFSET);
	case 3:
		return v3_to_v5((void *)buffer, (void *)buffer + COMPAT_OFFSET);
	case 4:
		return v4_to_v5((void *)buffer, (void *)buffer + COMPAT_OFFSET);
	}

	return -EINVAL;
}

static int v5_to_v2(struct clog_request *rq, struct clog_tfr_v2 *v2)
{
	LOG_ERROR("Unable to convert v5->v2");
	return -EINVAL;
}

static int v5_to_v3(struct clog_request *rq, struct clog_tfr_v3 *v3)
{
	LOG_ERROR("Unable to convert v5->v3");
	return -EINVAL;
}

static int v5_to_v4_copy_data(struct clog_request *rq, struct clog_tfr_v4 *v4)
{
	int data_size = 0;
	struct dm_ulog_request *u_rq = &rq->u_rq;
	uint32_t rq_type = le32_to_cpu(u_rq->request_type) & ~DM_CLOG_RESPONSE;

	if (le32_to_cpu(u_rq->request_type) & DM_CLOG_RESPONSE) {
		switch (rq_type) {
		case DM_CLOG_CTR:
		case DM_CLOG_DTR:
		case DM_CLOG_PRESUSPEND:
		case DM_CLOG_POSTSUSPEND:
		case DM_CLOG_RESUME:
		case DM_CLOG_FLUSH:
		case DM_CLOG_MARK_REGION:
		case DM_CLOG_CLEAR_REGION:
		case DM_CLOG_SET_REGION_SYNC:
		case DM_CLOG_CHECKPOINT_READY:
		case DM_CLOG_MEMBER_JOIN:
			/* No outbound data */
			break;

		case DM_CLOG_GET_REGION_SIZE:
			data_size = sizeof(uint64_t);
			if (v4) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)rq->u_rq.data;
				b64 = (uint64_t *)v4->data;
				*b64 = le64_to_cpu(*a64);
			}
			break;
		case DM_CLOG_IS_CLEAN:
		case DM_CLOG_IN_SYNC:
			data_size = sizeof(int);
			if (v4) {
				int *p;
				int64_t *p64;

				p = (int *)v4->data;
				p64 = (int64_t *)rq->u_rq.data;
				*p = (int)(le64_to_cpu(*p64));
			}
			break;
		case DM_CLOG_GET_RESYNC_WORK:
			data_size = sizeof(int32_t) +
				sizeof(uint32_t) + sizeof(uint64_t);
			if (v4) {
				int32_t *pi32;
				int64_t *pi64;
				uint64_t *pu64, u64;

				pi32 = (int32_t *)v4->data;
				pu64 = ((uint64_t *)v4->data) + 1;

				pi64 = (int64_t *)rq->u_rq.data;
				u64 = ((uint64_t *)rq->u_rq.data)[1];

				*pi32 = (int32_t)(le64_to_cpu(*pi64));
				*pu64 = (uint64_t)le64_to_cpu(u64);
			}
			break;
		case DM_CLOG_GET_SYNC_COUNT:
			data_size = sizeof(uint64_t);
			if (v4) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)rq->u_rq.data;
				b64 = (uint64_t *)v4->data;
				*b64 = le64_to_cpu(*a64);
			}
			break;			
		case DM_CLOG_STATUS_INFO:
		case DM_CLOG_STATUS_TABLE:
			data_size = u_rq->data_size;
			if (v4)
				memcpy(v4->data, rq->u_rq.data, data_size);
			break;
		case DM_CLOG_IS_REMOTE_RECOVERING:
			data_size = sizeof(int32_t) +
				sizeof(uint32_t) + sizeof(uint64_t);
			if (v4) {
				int32_t *pi32;
				uint64_t *pu64, u64;
				int64_t i64;

				i64 = ((int64_t *)rq->u_rq.data)[0];
				u64 = ((uint64_t *)rq->u_rq.data)[1];

				pi32 = (int32_t *)v4->data;
				pu64 = ((uint64_t *)v4->data) + 1;

				*pi32 = (int32_t)(le64_to_cpu(i64));
				*pu64 = le64_to_cpu(u64);
			}
			break;
		default:
			LOG_ERROR("Unknown request type, %u", rq_type);
			return -EINVAL;
		}
	} else {
		switch (rq_type) {
		case DM_CLOG_CTR:
		case DM_CLOG_DTR:
		case DM_CLOG_PRESUSPEND:
		case DM_CLOG_POSTSUSPEND:
		case DM_CLOG_RESUME:
		case DM_CLOG_GET_REGION_SIZE:
		case DM_CLOG_FLUSH:
		case DM_CLOG_GET_RESYNC_WORK:
		case DM_CLOG_GET_SYNC_COUNT:
		case DM_CLOG_STATUS_INFO:
		case DM_CLOG_STATUS_TABLE:
		case DM_CLOG_CHECKPOINT_READY:
		case DM_CLOG_MEMBER_JOIN:
			/* No incoming data */
			break;
		case DM_CLOG_IS_CLEAN:
		case DM_CLOG_IN_SYNC:
			data_size = sizeof(uint64_t);
			if (v4) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)rq->u_rq.data;
				b64 = (uint64_t *)v4->data;
				*b64 = le64_to_cpu(*a64);
			}
			break;
		case DM_CLOG_MARK_REGION:
		case DM_CLOG_CLEAR_REGION:
			data_size = le32_to_cpu(u_rq->data_size);
			if (v4) {
				int i, end = data_size/sizeof(uint64_t);
				uint64_t *a64, *b64;

				a64 = (uint64_t *)rq->u_rq.data;
				b64 = (uint64_t *)v4->data;
				for (i = 0; i < end; i++)
					b64[i] = le64_to_cpu(a64[i]);
			}
			break;
		case DM_CLOG_SET_REGION_SYNC:
			data_size = sizeof(uint64_t) +
				sizeof(uint32_t) + sizeof(int32_t);
			if (v4) {
				uint64_t u64, *pu64;
				int64_t i64;
				int32_t *pi32;

				pu64 = (uint64_t *)v4->data;
				pi32 = ((int32_t *)v4->data) + 3;

				u64 = ((uint64_t *)rq->u_rq.data)[0];
				i64 = ((int64_t *)rq->u_rq.data)[1];

				*pu64 = le64_to_cpu(u64);
				*pi32 = (uint32_t)(le64_to_cpu(i64));
			}
			break;
		case DM_CLOG_IS_REMOTE_RECOVERING:
			data_size = sizeof(uint64_t);
			if (v4) {
				uint64_t *a64, *b64;

				a64 = (uint64_t *)rq->u_rq.data;
				b64 = (uint64_t *)v4->data;
				*b64 = le64_to_cpu(*a64);
			}
			break;
		default:
			LOG_ERROR("Unknown request type, %u", rq_type);
			return -EINVAL;
		}
	}

	return data_size;
}

static int v5_to_v4(struct clog_request *rq, struct clog_tfr_v4 *v4)
{
	int size;
	struct dm_ulog_request *u_rq = &rq->u_rq;

	size = sizeof(*v4) + v5_to_v4_copy_data(rq, NULL);

	if (size > (COMPAT_SIZE - COMPAT_OFFSET)) {
		LOG_ERROR("Not enough space for compatibility data (v5->v4)");
		return -EINVAL;
	}

	v4->uuid_instance = (uint32_t)u_rq->luid;

	memcpy(v4->uuid, u_rq->uuid, DM_UUID_LEN);
	v4->error = u_rq->error;
	v4->seq = u_rq->seq;
	v4->request_type = u_rq->request_type;
	v4->data_size = u_rq->data_size;

	v4->originator = rq->originator;

	v5_to_v4_copy_data(rq, v4);

	return 0;
}

int v5_endian_from_wire(struct clog_request *rq)
{
	int size;
	struct dm_ulog_request *u_rq = &rq->u_rq;

	u_rq->error = le32_to_cpu(u_rq->error);
	u_rq->seq = le32_to_cpu(u_rq->seq);
	u_rq->request_type = le32_to_cpu(u_rq->request_type);
	u_rq->data_size = le64_to_cpu(u_rq->data_size);

	rq->originator = le32_to_cpu(rq->originator);

	size = sizeof(*rq) + u_rq->data_size;

	v5_data_endian_switch(rq, 0);

	return size;
}

int strip_compatibility_layer(void *from, void **to,
			      unsigned to_version)
{
	static int report = 1;
	uint64_t *vp = from;
	uint64_t version = le64_to_cpu(*vp);
	uint64_t unconverted_version = vp[1];

	/* We will modify the front of the area in place */
	*to = from;

	/* Already the version we need and same byte-order? */
	if ((version == to_version) && (unconverted_version == version))
		return 0;

	/* A different acceptable version? */
	if ((version > 1) && (version < 5)) {
		/* We can use the compatibility layer to convert */
		switch (to_version) {
		case 2:
			return v5_to_v2(from + COMPAT_OFFSET, from);
		case 3:
			return v5_to_v3(from + COMPAT_OFFSET, from);
		case 4:
			return v5_to_v4(from + COMPAT_OFFSET, from);
		case 5:
			*to = from + COMPAT_OFFSET;
			return 0;
		}
	}

	/* Is it upstream format? */
	if (version == 5) {
		*to = buffer;
		switch (to_version) {
		case 2:
			return v5_to_v2(from, (void *)buffer);
		case 3:
			return v5_to_v3(from, (void *)buffer);
		case 4:
			return v5_to_v4(from, (void *)buffer);
		case 5:
			return v5_endian_from_wire(from);
		}
		return -EINVAL;
	}		

	/*
	 * Is version unknown?  (From a release that didn't specify?)
	 * We must assume it is the same version - that the user is
	 * z-stream updating first.
	 *
	 * IOW, assume (version == to_version)
	 */
	if (report) {
		LOG_ERROR("Unknown version for communication struct - forced to assume version %d", to_version);
		report = 0;
	}

	return 0;
}
